#!/usr/bin/env python
"""Module with Sounding Reference Signal (SRS) related functions"""

from typing import Optional, cast

import numpy as np

from pyphysim.reference_signals.srs import UeSequence

from .root_sequence import RootSequence
from .zadoffchu import get_shifted_root_seq

__all__ = ['get_dmrs_seq', 'DmrsUeSequence']


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Module Functions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def get_dmrs_seq(root_seq: np.ndarray, n_cs: int) -> np.ndarray:
    """
    Get the shifted root sequence suitable as the DMRS sequence of a user.

    Parameters
    ----------
    root_seq : np.ndarray
        The root sequence to shift.
    n_cs : int
        The desired cyclic shift number. This should be an integer from
        0 to 11, where 0 will just return the base sequence, 1 gives the
        first shift, and so on.

    Returns
    -------
    np.ndarray
        The shifted root sequence.

    See Also
    --------
    .zadoffchu.get_shifted_root_seq, .srs.get_srs_seq
    """
    return get_shifted_root_seq(root_seq, n_cs, 12)


class DmrsUeSequence(UeSequence):
    """
    DMRS sequence of a single user.

    Parameters
    ----------
    root_seq : RootSequence
        The DMRS root sequence of the base station the user is
        associated to. This should be an object of the RootSequence
        class.
    n_cs : int
        The shift index of the user. This can be an integer from 0 to 11.
    cover_code : np.ndarray, optional
        Cover Code used by the UE. As an example, consider the cover code
        `np.array([1, -1])`. In that case, if the regular DMRS sequence
        (without the cover code) is `seq`, than the actual DMRS sequence
        with cover code will be a 2D numpy array equivalent with
        `seq_occ[0]==seq` and `seq_occ[1]==-seq`.
    normalize : bool
        True if the reference signal should be normalized. False otherwise.
    """
    def __init__(self,
                 root_seq: RootSequence,
                 n_cs: int,
                 cover_code: Optional[np.ndarray] = None,
                 normalize: bool = False) -> None:
        root_seq_array = root_seq.seq_array()
        user_seq_array = get_dmrs_seq(root_seq_array, n_cs)

        # Orthogonal Cover Code. This is stored in an attribute only for
        # visualization purposes, since the stored user_seq_array will
        # already include its effect.
        self._occ = cover_code

        if cover_code is not None:
            assert (isinstance(self._occ, np.ndarray))
            self._occ.flags.writeable = False
            user_seq_array = user_seq_array * cover_code[:, np.newaxis]

        super(DmrsUeSequence, self).__init__(root_seq,
                                             n_cs,
                                             user_seq_array,
                                             normalize=normalize)

    @property
    def cover_code(self) -> np.ndarray:
        """Return the cover code."""
        return self._occ

    @property
    def size(self) -> int:
        """
        Return the size of the reference signal sequence.

        Returns
        -------
        size : int
            The size of the user's reference signal sequence.
        """
        if self._occ is None:
            return cast(int, self._user_seq_array.shape[0])

        return cast(int, self._user_seq_array.shape[1])

    def __repr__(self) -> str:
        """
        Get the representation of the object.

        Returns
        -------
        str
            The representation of the object.
        """
        return "<{0}(root_index={1}, n_cs={2}, cover_code={3})>".format(
            self.__class__.__name__, self._root_index, self._n_cs, self._occ)



#!/usr/bin/env python
"""
Module with channel estimation implementations based on the reference signals in
this package.
"""

from typing import Union

import numpy as np

from .dmrs import DmrsUeSequence
from .srs import SrsUeSequence, UeSequence


class CazacBasedChannelEstimator:
    """
    Estimated the (uplink) channel based on CAZAC (Constant Amplitude Zero
    AutoCorrelation) reference sequences sent by one user (either SRS or
    DMRS).

    The estimation is performed according to the paper [Bertrand2011]_,
    where the received signal in the FREQUENCY DOMAIN is used by the
    estimator.

    Note that for SRS sequences usually a comb pattern is employed such
    that only half of the subcarriers is used to send pilot
    symbols. Therefore, an FFT in the during the estimation will
    effectively interpolate for the other subcarriers. This is controlled
    by the `size_multiplier` argument (default is 2 to accommodate comb
    pattern). If all subcarriers are used to send pilot symbols then set
    `size_multiplier` to 1.

    Parameters
    ----------
    ue_ref_seq : SrsUeSequence | DmrsUeSequence | np.ndarray
        The reference signal sequence.
    size_multiplier : int, optional
        Multiplication factor for the FFT to get the actual channel size.
        When using the comb pattern for SRS this should be 2 (default value),
        but for DMRS, which does not employ the comb pattern, this should be
        set to 1.

    Notes
    -----

    .. [Bertrand2011] Bertrand, Pierre, "Channel Gain Estimation from
       Sounding Reference Signal in LTE," Conference: Proceedings of the
       73rd IEEE Vehicular Technology Conference.
    """
    def __init__(self,
                 ue_ref_seq: Union[SrsUeSequence, DmrsUeSequence, np.ndarray],
                 size_multiplier: int = 2) -> None:
        # If ue_ref_seq is not an instance of UeSequence (or a subclass)
        # assume it is a numpy array.
        if isinstance(ue_ref_seq, UeSequence):
            self._normalized_ref_seq = ue_ref_seq.normalized
            ue_ref_seq = ue_ref_seq.seq_array()
        else:
            self._normalized_ref_seq = False

        self._ue_ref_sequence = ue_ref_seq
        self._size_multiplier = size_multiplier

    @property
    def ue_ref_seq(self) -> np.ndarray:
        """Get the sequence of the UE."""
        return self._ue_ref_sequence

    def estimate_channel_freq_domain(self, received_signal: np.ndarray,
                                     num_taps_to_keep: int) -> np.ndarray:
        """
        Estimate the channel based on the received signal.

        Parameters
        ----------
        received_signal : np.ndarray
            The received reference signal after being transmitted through the
            channel (in the frequency domain). If this is a 2D numpy array
            the first dimensions is assumed to be "receive antennas" while
            the second dimension are the received sequence elements. The
            number of elements in the received signal (per antenna) is
            equal to the channel size (number of subcarriers) divided by
            `size_multiplier`.
        num_taps_to_keep : int
            Number of taps (in delay domain) to keep. All taps from 0 to
            `num_taps_to_keep`-1 will be kept and all other taps will be
            zeroed before applying the FFT to get the channel response in
            the frequency domain.

        Returns
        -------
        freq_response : np.ndarray
            The channel frequency response. Note that for SRS sequences
            this will have twice as many elements as the sent SRS signal,
            since the SRS signal is sent every other subcarrier.
        """
        # Reference signal sequence
        r = self.ue_ref_seq

        if received_signal.ndim == 1:
            # First we multiply (element-wise) the received signal by the
            # conjugate of the reference signal sequence
            y = np.fft.ifft(np.conj(r) * received_signal, r.size)

            # The channel impulse response consists of the first
            # `num_taps_to_keep` elements in `y`.
            tilde_h = y[0:num_taps_to_keep + 1]
        elif received_signal.ndim == 2:
            # Case with multiple receive antennas
            y = np.fft.ifft(
                np.conj(r)[np.newaxis, :] * received_signal, r.size)

            # The channel impulse response consists of the first
            # `num_taps_to_keep` elements in `y`.
            tilde_h = y[:, 0:num_taps_to_keep + 1]
        else:
            raise ValueError(  # pragma: no cover
                "received_signal must have either one dimension ("
                "one receive antenna) or two dimensions (first "
                "dimension being the receive antenna dimension).")

        # Now we can apply the FFT to get the frequency response.
        # The number of subcarriers is twice the number of elements in the
        # SRS sequence due to the comb pattern
        Nsc = r.size
        tilde_H = np.fft.fft(tilde_h, self._size_multiplier * Nsc)

        if self._normalized_ref_seq is True:
            tilde_H *= Nsc

        return tilde_H


class CazacBasedWithOCCChannelEstimator(CazacBasedChannelEstimator):
    """
    Estimated the (uplink) channel based on CAZAC (Constant Amplitude Zero
    AutoCorrelation) reference sequences sent by one user including the
    Orthogonal Cover Code (OCC).

    With OCC the user will send reference signal in multiple time slots, in
    each slot multiplied with the respective OCC sequence element.

    Parameters
    ----------
    ue_ref_seq : DmrsUeSequence
        The reference signal sequence.
    """
    def __init__(self, ue_ref_seq: DmrsUeSequence) -> None:
        cover_code = ue_ref_seq.cover_code
        ue_ref_seq_array = ue_ref_seq.seq_array()
        reference_seq = ue_ref_seq_array[0] * cover_code[0]

        super(CazacBasedWithOCCChannelEstimator,
              self).__init__(reference_seq, size_multiplier=1)

        self._cover_code = cover_code
        self._normalized_ref_seq = ue_ref_seq.normalized

    @property
    def cover_code(self) -> np.ndarray:
        """Get the cover code of the UE."""
        return self._cover_code

    def estimate_channel_freq_domain(
            self,
            received_signal: np.ndarray,
            num_taps_to_keep: int,
            extra_dimension: bool = True) -> np.ndarray:
        """
        Estimate the channel based on the received signal with cover codes.

        Parameters
        ----------
        received_signal : np.ndarray
            The received reference signal after being transmitted through
            the channel (in the frequency domain).

            Dimension: Depend if there are multiple receive antennas and if
            `extra_dimension` is True or False. Let :math:`Nr` be the
            number of receive antennas, :math:`Ne` be the number of reference
            signal elements (reference signal size without cover code) and
            :math:`Nc` be the cover code size. The dimension of
            `received_signal` must match the table below.

            =================  =======================  ======================
                   /            extra_dimension: True   extra_dimension: False
            =================  =======================  ======================
            Single Antenna      Nc x Ne           (2D)   Ne * Nc          (1D)
            Multiple Antennas   Nr x Nc x Ne      (3D)   Nr x (Ne * Nc)   (2D)
            =================  =======================  ======================

        num_taps_to_keep : int
            Number of taps (in delay domain) to keep. All taps from 0 to
            `num_taps_to_keep`-1 will be kept and all other taps will be
            zeroed before applying the FFT to get the channel response in
            the frequency domain.
        extra_dimension : bool
            If True then the should be an extra dimension in
            `received_signal` corresponding to the cover code dimension. If
            False then the cover code is included in the dimension of the
            reference signal elements.

        Returns
        -------
        freq_response : np.ndarray
            The channel frequency response.
        """
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxx Add the extra dimension if it does not exist xxxxxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Create a view for the received signals. If extra_dimension is
        # false we will reshape this view to add a dimension for the cover
        # code
        r = received_signal.view()
        if extra_dimension is False:
            # Let's reorganize the received signal so that we have the
            # extra dimension
            if received_signal.ndim == 1:
                # Case with a single antenna. Cover code dimension will be
                # the first dimension.
                r.shape = (self.cover_code.size, -1)

            elif received_signal.ndim == 2:
                # Case with multiple antennas. Cover code dimension will be
                # the second dimension.
                num_antennas = r.shape[0]
                r.shape = (num_antennas, self.cover_code.size, -1)
            else:
                raise RuntimeError(
                    'Invalid dimension for received_signal: {0}'.format(
                        r.ndim))

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxxxxxxx Average over the cover code dimension xxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Now we can consider the case with the extra cover_code dimension
        if r.ndim == 2:
            # Apply the cover code
            r_mean = np.mean(r * self.cover_code[:, np.newaxis], axis=0)
        elif r.ndim == 3:
            r_mean = np.mean(r * self.cover_code[np.newaxis, :, np.newaxis],
                             axis=1)

        else:
            raise RuntimeError(
                'Invalid dimension for received_signal: {0}'.format(r.ndim))

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxxxxxxx Perform the estimation xxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Call the estimate_channel_freq_domain from the base class
        return super(CazacBasedWithOCCChannelEstimator, self).\
            estimate_channel_freq_domain(r_mean, num_taps_to_keep)



#!/usr/bin/env python
"""Package with communication related modules such as modulators, channels,
etc.

1. summary
2. extended summary
3. routine listings
4. see also
5. notes
6. references
7. examples

"""

# import blockdiagonalization
# import channels
# import mimo
# import modulators
# import ofdm
# import pathloss
# import waterfilling

from . import blockdiagonalization, waterfilling



#!/usr/bin/env python
"""
Module containing function related to serialization.
"""

import json
from typing import Any, Dict, List, Union

import numpy as np

Serializable = Union[np.ndarray, np.int32, np.int64, np.float32, np.float64,
                     np.float128, set]

# A type corresponding to the JSON representation of the object. For a lack of
# a better option we use Any
JsonRepresentation = Any


class NumpyOrSetEncoder(json.JSONEncoder):
    """
    JSON encoder for numpy arrays.

    Pass this class to json.dumps when converting a dictionary to json so
    that any field which with a numpy array as value will be properly
    converted.

    This encoder will also handle numpy scalars and the native python set
    types.

    When you need to convert the json representation back, use the
    `json_numpy_or_set_obj_hook` function.

    See Also
    --------
    json_numpy_or_set_obj_hook
    """
    def default(self, obj: Serializable) -> JsonRepresentation:
        """
        If input object is an ndarray it will be converted into a dict holding
        data, dtype, _is_numpy_array and shape.

        Parameters
        ----------
        obj : Serializable

        Returns
        -------
        Serialized Data
        """
        # Case for numpy arrays
        if isinstance(obj, np.ndarray):
            return {
                'data': obj.tolist(),
                'dtype': str(obj.dtype),
                '_is_numpy_array': True,
                'shape': obj.shape
            }
        # Case for numpy scalars
        if isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        if isinstance(obj, (np.float32, np.float64, np.float128)):
            return int(obj)

        # Case for built-in Python sets
        if isinstance(obj, set):
            return {'data': list(obj), '_is_set': True}

        # If it is not a numpy array we fall back to base class encoder
        return json.JSONEncoder(self, obj)  # type: ignore


def json_numpy_or_set_obj_hook(
        dct: Dict[str, JsonRepresentation]) -> Serializable:
    """
    Decodes a previously encoded numpy array.

    Parameters
    ----------
    dct : dict
        The JSON encoded numpy array.

    Returns
    -------
    np.ndarray | set | dict, optional
        The decoded numpy array or None if the encoded json data was not an
        encoded numpy array.

    See Also
    --------
    NumpyOrSetEncoder
    """
    if isinstance(dct, dict) and '_is_numpy_array' in dct:
        if dct['_is_numpy_array'] is True:
            data = dct['data']
            return np.array(data)

        raise ValueError(
            'Json representation contains the "_is_numpy_array" key '
            'indicating that the object should be a numpy array, but it '
            'was set to False, which is not valid.')
    if isinstance(dct, dict) and '_is_set' in dct:
        if dct['_is_set'] is True:
            data = dct['data']
            return set(data)

        raise ValueError(
            'Json representation contains the "_is_set" key '
            'indicating that the object should be python set, but it '
            'was set to False, which is not valid.')
    return dct


class JsonSerializable:
    """
    Base class for classes you want to be JSON serializable (convert
    to/from JSON).

    You can call the methods `to_json` and `from_json` methods (the later
    is a staticmethod).

    Note that a subclass must implement the `_to_dict` and `_from_dict` methods.
    """
    def _to_dict(self) -> Dict[str, Any]:
        """
        Convert the object to a dictionary representation.

        Returns
        -------
        dict
            The dictionary representation of the object.
        """
        raise NotImplementedError("Implement in a subclass")

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the object to a dictionary representation.

        Returns
        -------
        dict
            The dictionary representation of the object.
        """
        return self._to_dict()

    @staticmethod
    def _from_dict(d: Dict[str, Any]) -> Any:
        """
        Convert from a dictionary to an object.

        Parameters
        ----------
        d : dict
            The dictionary representing the object.

        Returns
        -------
        Result
            The converted object.
        """
        raise NotImplementedError("Implement in a subclass")

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Any:
        """
        Convert from a dictionary to an object.

        Parameters
        ----------
        d : dict
            The dictionary representing the Result.

        Returns
        -------
        Result
            The converted object.
        """
        return cls._from_dict(d)

    def to_json(self) -> JsonRepresentation:
        """
        Convert the object to JSON.

        Returns
        -------
        str
            JSON representation of the object.
        """
        return json.dumps(self._to_dict(), cls=NumpyOrSetEncoder)

    @classmethod
    def from_json(cls, data: JsonRepresentation) -> Any:
        """
        Convert a JSON representation of the object to an actual object.

        Parameters
        ----------
        data : str
            The JSON representation of the object.

        Returns
        -------
        any
            The actual object
        """
        d = json.loads(data, object_hook=json_numpy_or_set_obj_hook)
        return cls._from_dict(d)


# xxxxxxxxxx Test and Example Usage xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
if __name__ == '__main__':
    expected = np.arange(100, dtype=np.float)
    dumped = json.dumps(expected, cls=NumpyOrSetEncoder)
    result = json.loads(dumped, object_hook=json_numpy_or_set_obj_hook)
    print(type(result))
    print(result)
